/*
 * kgraft_patch_bsc1046206
 *
 * Fix for CVE-2017-8797, bsc#1046206
 *
 *  Upstream commit:
 *  f961e3f2acae94b727380c0b74e2d3954d0edf79 ("nfsd: encoders mustn't use unitialized values in error cases")
 *
 *  SLE12-SP2 commit:
 *  Fix arrived via 4.4.70, i.e. 8e4a191e636cebf57f4a0d60b15aa8b5fafd0715
 *
 *  SLE12(-SP1) commit:
 *  Not affected, pNFS has been introduced with 4.0 only.
 *
 *  Copyright (c) 2017 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/sunrpc/xdr.h>
#include <linux/nfsd/debug.h>
#include <linux/nfsd/nfsfh.h>
#include <linux/slab.h>

#if !IS_MODULE(CONFIG_NFSD)
#error "KGR patch supports only CONFIG_NFSD=m."
#endif

#if !IS_MODULE(CONFIG_SUNRPC)
#error "KGR patch supports only CONFIG_SUNRPC=m."
#endif

#define KGR_PATCHED_MODULE "nfsd"


/* from patched fs/nfsd/nfs4xdr.c, needed for dprintk() */
#define NFSDDBG_FACILITY		NFSDDBG_XDR

/*
 * From fs/nfsd/nfsd.h. Be careful to avoid a dependency to the
 * nfsd_debug exported from sunrpc.
 */
#undef ifdebug
#ifdef CONFIG_SUNRPC_DEBUG
# define ifdebug(flag)		if (*kgr_nfsd_debug & NFSDDBG_##flag)
#else
# define ifdebug(flag)		if (0)
#endif

/* provided from net/sunrpc/sysctl.c, needed by ifdebug #define above */
#ifdef CONFIG_SUNRPC_DEBUG
static unsigned int *kgr_nfsd_debug;
#endif

/* from fs/nfsd/nfsd.h */
#define kgr_nfserr_toosmall	cpu_to_be32(NFSERR_TOOSMALL)
#define kgr_nfserr_resource	cpu_to_be32(NFSERR_RESOURCE)

/* only forward declaration needed */
struct nfs4_stateowner;
struct nfs4_client;
struct nfsd4_session;
struct nfsd4_slot;
struct nfsd4_layoutcommit;

/* from fs/nfsd/nfsfh.h */
typedef struct svc_fh {
	struct knfsd_fh		fh_handle;	/* FH data */
	int			fh_maxsize;	/* max size for fh_handle */
	struct dentry *		fh_dentry;	/* validated dentry */
	struct svc_export *	fh_export;	/* export pointer */

	bool			fh_locked;	/* inode locked by us */
	bool			fh_want_write;	/* remount protection taken */

#ifdef CONFIG_NFSD_V3
	bool			fh_post_saved;	/* post-op attrs saved */
	bool			fh_pre_saved;	/* pre-op attrs saved */

	/* Pre-op attributes saved during fh_lock */
	__u64			fh_pre_size;	/* size before operation */
	struct timespec		fh_pre_mtime;	/* mtime before oper */
	struct timespec		fh_pre_ctime;	/* ctime before oper */
	/*
	 * pre-op nfsv4 change attr: note must check IS_I_VERSION(inode)
	 *  to find out if it is valid.
	 */
	u64			fh_pre_change;

	/* Post-op attributes saved in fh_unlock */
	struct kstat		fh_post_attr;	/* full attrs after operation */
	u64			fh_post_change; /* nfsv4 change; see above */
#endif /* CONFIG_NFSD_V3 */

} svc_fh;

/* from fs/nfsd/state.h */
typedef struct {
	u32             cl_boot;
	u32             cl_id;
} clientid_t;

typedef struct {
	clientid_t	so_clid;
	u32		so_id;
} stateid_opaque_t;

typedef struct {
	u32                     si_generation;
	stateid_opaque_t        si_opaque;
} stateid_t;

/* from fs/nfsd/xdr4.h */
struct nfsd4_compound_state {
	struct svc_fh		current_fh;
	struct svc_fh		save_fh;
	struct nfs4_stateowner	*replay_owner;
	struct nfs4_client	*clp;
	/* For sessions DRC */
	struct nfsd4_session	*session;
	struct nfsd4_slot	*slot;
	int			data_offset;
	size_t			iovlen;
	u32			minorversion;
	__be32			status;
	stateid_t	current_stateid;
	stateid_t	save_stateid;
	/* to indicate current and saved state id presents */
	u32		sid_flags;
};

struct nfsd4_compoundres {
	/* scratch variables for XDR encode */
	struct xdr_stream		xdr;
	struct svc_rqst *		rqstp;

	u32				taglen;
	char *				tag;
	u32				opcnt;
	__be32 *			tagp; /* tag, opcount encode location */
	struct nfsd4_compound_state	cstate;
};

struct nfsd4_deviceid {
	u64			fsid_idx;
	u32			generation;
	u32			pad;
};

struct nfsd4_getdeviceinfo {
	struct nfsd4_deviceid	gd_devid;	/* request */
	u32			gd_layout_type;	/* request */
	u32			gd_maxcount;	/* request */
	u32			gd_notify_types;/* request - response */
	void			*gd_device;	/* response */
};

struct nfsd4_layout_seg {
	u32			iomode;
	u64			offset;
	u64			length;
};

struct nfsd4_layoutget {
	u64			lg_minlength;	/* request */
	u32			lg_signal;	/* request */
	u32			lg_layout_type;	/* request */
	u32			lg_maxcount;	/* request */
	stateid_t		lg_sid;		/* request/response */
	struct nfsd4_layout_seg	lg_seg;		/* request/response */
	void			*lg_content;	/* response */
};

/* from fs/nfsd/pnfs.h */
struct nfsd4_layout_ops {
	u32		notify_types;

	__be32 (*proc_getdeviceinfo)(struct super_block *sb,
			struct nfsd4_getdeviceinfo *gdevp);
	__be32 (*encode_getdeviceinfo)(struct xdr_stream *xdr,
			struct nfsd4_getdeviceinfo *gdevp);

	__be32 (*proc_layoutget)(struct inode *, const struct svc_fh *fhp,
			struct nfsd4_layoutget *lgp);
	__be32 (*encode_layoutget)(struct xdr_stream *,
			struct nfsd4_layoutget *lgp);

	__be32 (*proc_layoutcommit)(struct inode *inode,
			struct nfsd4_layoutcommit *lcp);
};

/*
 * Original declaration is in fs/nfsd/nfs4layouts.c:
 * const struct nfsd4_layout_ops *nfsd4_layout_ops[LAYOUT_TYPE_MAX]
 */
static const struct nfsd4_layout_ops* (*kgr_nfsd4_layout_ops)[];

static __be32* (*kgr_xdr_encode_opaque_fixed)(__be32 *p, const void *ptr,
					unsigned int nbytes);
static __be32* (*kgr_xdr_reserve_space)(struct xdr_stream *xdr, size_t nbytes);
static void (*kgr_xdr_truncate_encode)(struct xdr_stream *xdr, size_t len);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "sunrpc:nfsd_debug", (void *)&kgr_nfsd_debug },
	{ "nfsd:nfsd4_layout_ops", (void *)&kgr_nfsd4_layout_ops },
	{ "sunrpc:xdr_encode_opaque_fixed", (void *)&kgr_xdr_encode_opaque_fixed },
	{ "sunrpc:xdr_reserve_space", (void *)&kgr_xdr_reserve_space },
	{ "sunrpc:xdr_truncate_encode", (void *)&kgr_xdr_truncate_encode },
};

/* patched */
__be32 kgr_nfsd4_encode_getdeviceinfo(struct nfsd4_compoundres *resp,
				__be32 nfserr, struct nfsd4_getdeviceinfo *gdev)
{
	struct xdr_stream *xdr = &resp->xdr;
	/*
	 * Fix CVE-2017-8797
	 * -2 lines, +1 line
	 */
	const struct nfsd4_layout_ops *ops;
	u32 starting_len = xdr->buf->len, needed_len;
	__be32 *p;

	dprintk("%s: err %d\n", __func__, nfserr);
	if (nfserr)
		goto out;

	nfserr = kgr_nfserr_resource;
	p = kgr_xdr_reserve_space(xdr, 4);
	if (!p)
		goto out;

	*p++ = cpu_to_be32(gdev->gd_layout_type);

	/* If maxcount is 0 then just update notifications */
	if (gdev->gd_maxcount != 0) {
		/*
		 * Fix CVE-2017-8797
		 * +1 line
		 */
		ops = (*kgr_nfsd4_layout_ops)[gdev->gd_layout_type];
		nfserr = ops->encode_getdeviceinfo(xdr, gdev);
		if (nfserr) {
			/*
			 * We don't bother to burden the layout drivers with
			 * enforcing gd_maxcount, just tell the client to
			 * come back with a bigger buffer if it's not enough.
			 */
			if (xdr->buf->len + 4 > gdev->gd_maxcount)
				goto toosmall;
			goto out;
		}
	}

	nfserr = kgr_nfserr_resource;
	if (gdev->gd_notify_types) {
		p = kgr_xdr_reserve_space(xdr, 4 + 4);
		if (!p)
			goto out;
		*p++ = cpu_to_be32(1);			/* bitmap length */
		*p++ = cpu_to_be32(gdev->gd_notify_types);
	} else {
		p = kgr_xdr_reserve_space(xdr, 4);
		if (!p)
			goto out;
		*p++ = 0;
	}

	nfserr = 0;
out:
	kfree(gdev->gd_device);
	dprintk("%s: done: %d\n", __func__, be32_to_cpu(nfserr));
	return nfserr;

toosmall:
	dprintk("%s: maxcount too small\n", __func__);
	needed_len = xdr->buf->len + 4 /* notifications */;
	kgr_xdr_truncate_encode(xdr, starting_len);
	p = kgr_xdr_reserve_space(xdr, 4);
	if (!p) {
		nfserr = kgr_nfserr_resource;
	} else {
		*p++ = cpu_to_be32(needed_len);
		nfserr = kgr_nfserr_toosmall;
	}
	goto out;
}

/* patched */
__be32 kgr_nfsd4_encode_layoutget(struct nfsd4_compoundres *resp,
				__be32 nfserr, struct nfsd4_layoutget *lgp)
{
	struct xdr_stream *xdr = &resp->xdr;
	/*
	 * Fix CVE-2017-8797
	 * -2 lines, +1 line
	 */
	const struct nfsd4_layout_ops *ops;
	__be32 *p;

	dprintk("%s: err %d\n", __func__, nfserr);
	if (nfserr)
		goto out;

	nfserr = kgr_nfserr_resource;
	p = kgr_xdr_reserve_space(xdr, 36 + sizeof(stateid_opaque_t));
	if (!p)
		goto out;

	*p++ = cpu_to_be32(1);	/* we always set return-on-close */
	*p++ = cpu_to_be32(lgp->lg_sid.si_generation);
	p = kgr_xdr_encode_opaque_fixed(p, &lgp->lg_sid.si_opaque,
					sizeof(stateid_opaque_t));

	*p++ = cpu_to_be32(1);	/* we always return a single layout */
	p = xdr_encode_hyper(p, lgp->lg_seg.offset);
	p = xdr_encode_hyper(p, lgp->lg_seg.length);
	*p++ = cpu_to_be32(lgp->lg_seg.iomode);
	*p++ = cpu_to_be32(lgp->lg_layout_type);

	/*
	 * Fix CVE-2017-8797
	 * +1 line
	 */
	ops = (*kgr_nfsd4_layout_ops)[lgp->lg_layout_type];
	nfserr = ops->encode_layoutget(xdr, lgp);
out:
	kfree(lgp->lg_content);
	return nfserr;
}

static int kgr_patch_bsc1046206_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_bsc1046206_module_notify(struct notifier_block *nb,
	unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = kgr_patch_bsc1046206_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1046206_module_nb = {
	.notifier_call = kgr_patch_bsc1046206_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1046206_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = kgr_patch_bsc1046206_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1046206_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1046206_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1046206_module_nb);
}
